@@ -1,12 +1,11 @@
 module org.graalvm.nativeimage.pointsto {
     requires java.management;
+    requires jdk.graal.compiler;
     requires jdk.internal.vm.ci;
-    /* transitive */ requires jdk.internal.vm.compiler;
     requires jdk.management;
-    /* transitive */ requires org.graalvm.collections;
-    /* transitive */ requires org.graalvm.nativeimage;
-    /* transitive */ requires org.graalvm.nativeimage.base;
-    /* transitive */ requires org.graalvm.truffle.compiler;
+    requires org.graalvm.collections;
+    requires org.graalvm.nativeimage;
+    requires org.graalvm.nativeimage.base;
     /* transitive */ requires org.graalvm.word;
 
     exports com.oracle.graal.pointsto;
@@ -27,7 +26,7 @@ module org.graalvm.nativeimage.pointsto {
     exports com.oracle.graal.pointsto.typestate;
     exports com.oracle.graal.pointsto.util;
 
-    provides OptionDescriptors with StrengthenGraphs_OptionDescriptors, PointstoOptions_OptionDescriptors, AnalysisReportsOptions_OptionDescriptors, HeapSnapshotVerifier_OptionDescriptors, InlineBeforeAnalysis_OptionDescriptors;
+    provides OptionDescriptors with AnalysisReportsOptions_OptionDescriptors, StrengthenGraphs_OptionDescriptors, PointstoOptions_OptionDescriptors, HeapSnapshotVerifier_OptionDescriptors, InlineBeforeAnalysis_OptionDescriptors;
 
 }
 
